## (C) 2019 Mathematikinformation Nr. 72
## http://mathematikinformation.info/
##
## Lei Baozhen, Beijing Union University
## Harald Loewe, Technische Universitaet Braunschweig
## Numerische Verfahren zur Loesung von Differentialgleichungen
##
##############################################################
### Euler-Heun-Verfahren mit Schrittweitensteuerung        ###
##############################################################
##
## Ver. 2019-12-15
## Kodierung in UTF-8; Octave-Version 5.10
##
## Aufruf mit ODE_PZSW (f, t0, T, x0, opt)
## f = f(t,x) = rechte Seite der Differentialgleichung / des Systems
## t0 = Intervall-Anfang, T = Intervall-Ende
## x0 = x(t0) = Anfangswert
## opt = Optionen (optionales Argument)
## Syntax: {opt_1, Zahl_1, ..., opt_n, Zahl_n}
## Dabei ist opt_* eine der folgenden Strings:
##    'alpha'   = akzeptabler lokaler Fehler
##    'epsilon' = minimale Schrittweite
##    'sigma'   = Sicherheitsfaktor fuer Schrittweitenerhoehung
##    'dt'      = anfaengliche Schrittweite
## opt_1 wird auf Zahl_1 gesetzt, opt_2 auf Zahl_2 usw.
## Weglassen der Option opt entspricht
## {'alpha', 10^-2, 'epsilon', 10^-8, 'sigma', 0.8, 'dt', 1}
##
## Rueckgabe: (k+1)*N-Matrix M, wo 
##            k = Anzahl der gesuchten Funktionen des Systems
#             N = Anzahl der berechneten Werte
## M(1,:) = Liste der Stuetzstellen tk
## M(2,:) = Liste der Werte der ersten Funktion
## M(3,:) = Liste der Werte der zweiten Funktion
## usw.
## 
## This program is free software: you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <https://www.gnu.org/licenses/>.


function retval = ODE_EHvsPZ (f, t0, T, x0, opt = {})
  ## Standardwerte
  alp = 10^-2;  # akzeptabler lokaler Fehler
  dt = 1;       # anfaengliche Schrittweite
  sig = 0.8;    # Sicherheitsfaktor fuer Schrittweitenvergroesserung
  eps = 10^-8;  # minimale Schrittweite
  ## Ueberschreiben der Standardwerte
  for lauf = 1:(length(opt)/2)
    switch (opt{2*lauf-1})
      case 'alpha'
        alp = opt{2*lauf};
      case 'dt'
        dt = opt{2*lauf};
      case 'sigma'
        sig = opt{2*lauf};
      case 'epsilon'
        eps = opt{2*lauf};
      otherwise
        printf('Warnung! Unbekannte Option\n')
    endswitch
  endfor
  ## Ende der Zuweisung der Parameter fuer die Schrittweitensteuerung
  
  ## Initialisierung des Verfahrens:
  tk   = t0;            # aktuelle Stuetzstelle
  xk   = x0;            # Aktueller Wert
  ttt  = [t0];          # Liste der Stuetzstellen zu Beginn
  xxx  = [x0];          # Liste der x_k zu Beginn
  ## Aufbau der Listen per while-Schleife 
  ## Eine Initialisierung fehlt noch:
  q2 = f(t0,x0);
  while tk + dt <= T
    er = 2*alp;     # sonst laeuft die naechste Schleife leer
    while er > alp; # wir sind nicht zufrieden
      ## Berechne q1 und q2 fuer das Euler-Heun-Verfahren
      ## Eigentlich muessen wir rechnen:
      ## q1 = f(tk,xk) = f(tk_alt+dt_alt,xk_alt+dt_alt*q1_alt)
      ## Beobachtung: q1 haben wir bereits im letzten Schritt
      ## als das dort zustaendige q2 berechnet!
      q1 = q2;
      ## Polygonzugverfahren; diesen Wert brauchen wir mehrfach
      xP = xk + dt*q1;
      q2 = f(tk+dt,xP);
      ## Eigentlich geht es jetzt so weiter:
      ## Schaetze x(tk+dt) mit dem Polygonzugverfahren
      ## xa = xk + dt*q1;
      ## Schaetze x(tk+dt) mit dem Euler-Heun-Verfahren
      ## xb = xk + 0.5*dt*(q1+q2);
      ## Schaetze den Fehler nach Gleichung (34)
      ## er = norm(xb-xa);
      ## Wir sehen aber: xb - xa = 0.5*(q2-q1), also:
      er = 0.5*dt*norm(q2-q1);
      ## Fuer Systeme musste hier "norm" statt "abs" genommen werden
      if er > alp, dt = dt/2; end;  # Schrittweite halbieren!
      if dt < eps, error('Schrittweitenfehler'); return; end;
    end;
    ## Die letzte Schrittweite war akzeptabel; wir uebernehmen:
    tk  = tk + dt;
    ## Neues xk aus dem Polygonzugverfahren!
    xk  = xP;
    ## Update der Listen
    ttt = [ttt; tk];  # Werte an die Liste anhaengen
    xxx = [xxx; xk];  # dto.
    ## Schrittweite fuer das naechste Mal raten
    dt  = min([T-tk, sig*sqrt(alp/er)*dt]);
    if dt < eps      # Intervallende erreicht
      break;
    endif
  end;
  
  ## Rueckgabe der Ergebnisse als (k+1)*N-Matrix
  retval = transpose([ttt, xxx]);
endfunction
