## (C) 2019 Mathematikinformation Nr. 72
## http://mathematikinformation.info/
##
## Lei Baozhen, Beijing Union University
## Harald Loewe, Technische Universitaet Braunschweig
## Numerische Verfahren zur Loesung von Differentialgleichungen
##
##############################################################
### Eulersches Polgongzugverfahren                         ###
##############################################################
##
## Ver. 2019-12-15
## Kodierung in UTF-8; Octave-Version 5.10
##
## Aufruf mit ODE_PZ (f, t0, T, x0, dt [, xmax])
## f = f(t,x) = rechte Seite der Differentialgleichung / des Systems
## t0 = Intervall-Anfang, T = Intervall-Ende
## x0 = x(t0) = Anfangswert
## dt = Schrittweite
## xmax = Schranke fuer Werte |xk|
##  optionales Argument; Standardwert ist 10^36
## Rueckgabe: (k+1)*N-Matrix M, 
##             wo N = Anzahl der Schritte 
##             und k = Anzahl der gesuchten Funktionen des Systems
## M(1,:) = Liste der Stuetzstellen tk
## M(2,:) = Liste der Werte fuer die erste Funktion
## M(2,:) = Liste der Werte fuer die zweite Funktion
## usw.
## 
## This program is free software: you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <https://www.gnu.org/licenses/>.


function retval = ODE_PZ (f, t0, T, x0, dt, xmax = 10^36)
  ## Berechnen der Schrittweite
  ## Die Liste der Stuetzstellen sowie der Werte werden hier stueckweise 
  ## aufgebaut. Wir beginnen mit leeren Listen.
  ttt      = [];            # Liste der Stuetzstellen t_k
  xxx      = [];            # Liste der Werte x_k
  ## Anfangswerte
  last_x   = x0;
  last_t   = t0;
  while [max(abs(last_x)) <= xmax, last_t <= T] 
        ## Abbruchbedingung: 
        ## |x_j| > xmax fuer eine Komponenten x_j von x
        ## oder t > T
    ttt    = [ttt; last_t]; # Fortschreiben der Liste ttt
    xxx    = [xxx; last_x]; # Fortschreiben der Liste xxx
    last_x = last_x + dt * f(last_t,last_x); # Berechnung von x_(k+1)
    last_t = last_t + dt;                    # Naechste Stuetzstelle
  endwhile;
  ## Nachbehandlung...
  ttt_end  = ttt(length(ttt));      # Letzter Wert in ttt
  xxx_end  = xxx(size(xxx)(1),:);   # Letzter Wert in xxx
  if max(abs(last_x)) > xmax        # Zahlbereichsueberschreitung
    printf(["Werte nach t = ",num2str(ttt_end)," zu gross!\n"])
  elseif ttt_end < T                # Intervallende noch nicht erreicht;
                                    # schaetze x(T)
    xxx_end = xxx_end + (T-ttt_end) * f(ttt_end,xxx_end);
    ttt_end = T;
    if max(abs(xxx_end)) <= xmax     # Letzten Wert hinzunehmen
      ttt = [ttt; ttt_end];
      xxx = [xxx; xxx_end];
    else                            # Zahlbereichsueberschreitung
      printf(["Werte nach t = ",num2str(ttt_end)," zu gross!\n"])
    endif
  else                              # Alles in Ordnung; nichts zu tun.
  endif
  
  ## Rueckgabe der Ergebnisse als (k+1)*N-Matrix,
  ## wobei k die Anzahl der gesuchten Funktionen bezeichnet
  retval = transpose([ttt xxx]);  
endfunction
